"""The `*Repo` abstract types which implement the "Repository" pattern."""

from __future__ import annotations

import abc
from typing import Generic, TypeVar

from eris import ErisResult, Err, Ok


K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


class BasicRepo(Generic[K, V], abc.ABC):
    """The simplest possible Repository type."""

    @abc.abstractmethod
    def add(self, item: V, /, *, key: K = None) -> ErisResult[K]:
        """Add a new `item` to the repo and associsate it with `key`."""

    @abc.abstractmethod
    def get(self, key: K) -> ErisResult[V | None]:
        """Retrieve an item from the repo by key."""


class Repo(BasicRepo[K, V], Generic[K, V], abc.ABC):
    """A full-featured Repository

    Adds the ability to update, delete, and list all items ontop of the
    BasicRepo type.
    """

    @abc.abstractmethod
    def remove(self, key: K) -> ErisResult[V | None]:
        """Remove an item from the repo by key."""

    def update(self, key: K, item: V, /) -> ErisResult[V]:
        """Update an item by key."""
        old_item = self.remove(key).unwrap()
        if old_item is None:
            return Err(f"Old Todo with this ID does not exist. | id={key}")

        self.add(item, key=key).unwrap()

        return Ok(old_item)

    @abc.abstractmethod
    def all(self) -> ErisResult[list[V]]:
        """Retrieve all items stored in this repo."""


class TaggedRepo(Repo[K, V], Generic[K, V, T], abc.ABC):
    """A Repository that is aware of some kind of "tags".

    Adds the ability to retrieve / delete a group of objects based off of some
    arbitrary "tag" type.

    NOTE: In general, K can be expected to be a primitive type, whereas T is
      often a custom user-defined type.
    """

    @abc.abstractmethod
    def get_by_tag(self, tag: T) -> ErisResult[list[V]]:
        """Retrieve a group of items that meet the given tag's criteria."""

    @abc.abstractmethod
    def remove_by_tag(self, tag: T) -> ErisResult[list[V]]:
        """Remove a group of items that meet the given tag's criteria."""



"""Tests for the potoroo package."""

from __future__ import annotations

from eris import ErisResult, Ok

from potoroo import Repo, TaggedRepo


class FakeDB(Repo[int, str]):
    """Fake database."""

    def __init__(self) -> None:
        self._keys = list(range(100))
        self._db: dict[int, str] = {}

    def add(self, some_item: str, /, *, key: int = None) -> ErisResult[int]:
        """Fake add."""
        key = self._keys.pop(0)
        self._db[key] = some_item
        return Ok(key)

    def get(self, key: int) -> ErisResult[str | None]:
        """Fake get."""
        return Ok(self._db[key])

    def remove(self, key: int) -> ErisResult[str | None]:
        """Fake remove."""
        return Ok(self._db.pop(key))

    def update(self, key: int, some_item: str, /) -> ErisResult[str]:
        """Fake update."""
        self._db[key] = some_item
        return Ok(some_item)

    def all(self) -> ErisResult[list[str]]:
        """Fake all."""
        return Ok(sorted(self._db.values()))


class FakeTaggedDB(FakeDB, TaggedRepo[int, str, str]):
    """Fake tagged database."""

    def get_by_tag(self, tag: str) -> ErisResult[list[str]]:
        """Fake get_by_tag."""
        return Ok([v for v in self._db.values() if tag in v])

    def remove_by_tag(self, tag: str) -> ErisResult[list[str]]:
        """Fake remove_by_tag."""
        res: list[str] = []
        for k, v in dict(self._db).items():
            if tag in v:
                res.append(self._db.pop(k))
        return Ok(res)


def test_repo() -> None:
    """Test the Repo type."""
    db = FakeDB()
    foo_idx = db.add("foo").unwrap()
    assert db.get(foo_idx).unwrap() == "foo"
    assert db.update(foo_idx, "bar").unwrap() == "bar"
    assert db.remove(foo_idx).unwrap() == "bar"


def test_tagged_repo() -> None:
    """Test the TaggedRepo type."""
    db = FakeTaggedDB()
    foo_idx = db.add("foo").unwrap()
    db.add("bar").unwrap()
    db.add("baz").unwrap()

    assert db.get(foo_idx).unwrap() == "foo"
    assert db.get_by_tag("f").unwrap() == ["foo"]
    assert db.all().unwrap() == ["bar", "baz", "foo"]
    assert db.remove_by_tag("b").unwrap() == ["bar", "baz"]



"""Contains the `UnitOfWork` abstract type."""

from __future__ import annotations

import abc
from types import TracebackType
from typing import Generic, Type, TypeVar

from ._repo import BasicRepo


R = TypeVar("R", bound=BasicRepo)
U = TypeVar("U", bound="UnitOfWork")


class UnitOfWork(Generic[R], abc.ABC):
    """Implements the 'Unit-of-Work' persistance pattern."""

    @property
    @abc.abstractmethod
    def repo(self) -> R:
        """The Repo type this UoW is associated with."""

    @abc.abstractmethod
    def __enter__(self: U) -> U:
        """Called before entering the UoW context manager."""

    @abc.abstractmethod
    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Called before leaving the UoW context manager."""

    @abc.abstractmethod
    def commit(self) -> None:
        """Commits our changes to `self.repo`."""

    @abc.abstractmethod
    def rollback(self) -> None:
        """Reverts our changes to `self.repo`."""



"""Python implementations of the Repository and UnitOfWork abstractions."""

import logging as _logging

from ._repo import BasicRepo, Repo, TaggedRepo
from ._uow import UnitOfWork


__all__ = ["BasicRepo", "Repo", "TaggedRepo", "UnitOfWork"]

__author__ = "Bryan M Bugyi"
__email__ = "bryanbugyi34@gmail.com"
__version__ = "0.4.0"

_logging.getLogger(__name__).addHandler(_logging.NullHandler())
